@@ -1,4 +1,9 @@
 import jinja2
+try:
+    import jinja2.ext.autoescape
+    jinja2_autoescape_builtin = False
+except ImportError:
+    jinja2_autoescape_builtin = True
 import flask
 import httpauth
 import dulwich.web
@@ -7,24 +12,24 @@ from klaus import views, utils
 from klaus.repo import FancyRepo, InvalidRepo
 
 
-KLAUS_VERSION = utils.guess_git_revision() or '1.5.2'
+KLAUS_VERSION = utils.guess_git_revision() or "2.0.1"
 
 
 class Klaus(flask.Flask):
     jinja_options = {
-        'extensions': ['jinja2.ext.autoescape'],
-        'undefined': jinja2.StrictUndefined
+        "extensions": [] if jinja2_autoescape_builtin else ["jinja2.ext.autoescape"],
+        "undefined": jinja2.StrictUndefined,
     }
 
-    def __init__(self, repo_paths, site_name, use_smarthttp, ctags_policy='none'):
+    def __init__(self, repo_paths, site_name, use_smarthttp, ctags_policy="none"):
         """(See `make_app` for parameter descriptions.)"""
         self.site_name = site_name
         self.use_smarthttp = use_smarthttp
         self.ctags_policy = ctags_policy
 
         valid_repos, invalid_repos = self.load_repos(repo_paths)
-        self.valid_repos = {repo.name: repo for repo in valid_repos}
-        self.invalid_repos = {repo.name: repo for repo in invalid_repos}
+        self.valid_repos = {repo.namespaced_name: repo for repo in valid_repos}
+        self.invalid_repos = {repo.namespaced_name: repo for repo in invalid_repos}
 
         flask.Flask.__init__(self, __name__)
 
@@ -34,22 +39,23 @@ class Klaus(flask.Flask):
         """Called by Flask.__init__"""
         env = super(Klaus, self).create_jinja_environment()
         for func in [
-            'force_unicode',
-            'timesince',
-            'shorten_sha1',
-            'shorten_message',
-            'extract_author_name',
-            'formattimestamp',
+            "force_unicode",
+            "timesince",
+            "shorten_sha1",
+            "shorten_message",
+            "extract_author_name",
+            "formattimestamp",
         ]:
             env.filters[func] = getattr(utils, func)
 
-        env.globals['KLAUS_VERSION'] = KLAUS_VERSION
-        env.globals['USE_SMARTHTTP'] = self.use_smarthttp
-        env.globals['SITE_NAME'] = self.site_name
+        env.globals["KLAUS_VERSION"] = KLAUS_VERSION
+        env.globals["USE_SMARTHTTP"] = self.use_smarthttp
+        env.globals["SITE_NAME"] = self.site_name
 
         return env
 
     def setup_routes(self):
+        # fmt: off
         for endpoint, rule in [
             ('repo_list',   '/'),
             ('robots_txt',  '/robots.txt/'),
@@ -71,13 +77,18 @@ class Klaus(flask.Flask):
             ('download',    '/<repo>/tarball/<path:rev>/'),
         ]:
             self.add_url_rule(rule, view_func=getattr(views, endpoint))
+            if "<repo>" in rule:
+                self.add_url_rule(
+                    "/~<namespace>" + rule, view_func=getattr(views, endpoint)
+                )
+        # fmt: on
 
     def should_use_ctags(self, git_repo, git_commit):
-        if self.ctags_policy == 'none':
+        if self.ctags_policy == "none":
             return False
-        elif self.ctags_policy == 'ALL':
+        elif self.ctags_policy == "ALL":
             return True
-        elif self.ctags_policy == 'tags-and-branches':
+        elif self.ctags_policy == "tags-and-branches":
             return git_commit.id in git_repo.get_tag_and_branch_shas()
         else:
             raise ValueError("Unknown ctags policy %r" % self.ctags_policy)
@@ -85,22 +96,36 @@ class Klaus(flask.Flask):
     def load_repos(self, repo_paths):
         valid_repos = []
         invalid_repos = []
-        for path in repo_paths:
-            try:
-                valid_repos.append(FancyRepo(path))
-            except NotGitRepository:
-                invalid_repos.append(InvalidRepo(path))
+        for namespace, paths in repo_paths.items():
+            for path in paths:
+                try:
+                    valid_repos.append(FancyRepo(path, namespace))
+                except NotGitRepository:
+                    invalid_repos.append(InvalidRepo(path, namespace))
         return valid_repos, invalid_repos
 
 
-def make_app(repo_paths, site_name, use_smarthttp=False, htdigest_file=None,
-             require_browser_auth=False, disable_push=False, unauthenticated_push=False,
-             ctags_policy='none'):
+def make_app(
+    repo_paths,
+    site_name,
+    use_smarthttp=False,
+    htdigest_file=None,
+    require_browser_auth=False,
+    disable_push=False,
+    unauthenticated_push=False,
+    ctags_policy="none",
+):
     """
     Returns a WSGI app with all the features (smarthttp, authentication)
     already patched in.
-
-    :param repo_paths: List of paths of repositories to serve.
+    :param repo_paths: Repositories to serve. This can either be a list of paths
+        or dictionary of the following form:
+            {
+                "namespace1": [list of paths of repositories],
+                "namespace2": [list of paths of repositories],
+                ...
+                None: [list of paths of repositories without namespace]
+            }
     :param site_name: Name of the Web site (e.g. "John Doe's Git Repositories")
     :param use_smarthttp: Enable Git Smart HTTP mode, which makes it possible to
         pull from the served repositories. If `htdigest_file` is set as well,
@@ -125,9 +150,16 @@ def make_app(repo_paths, site_name, use_smarthttp=False, htdigest_file=None,
         if disable_push:
             raise ValueError("'unauthenticated_push' set with 'disable_push'")
         if require_browser_auth:
-            raise ValueError("Incompatible options 'unauthenticated_push' and 'require_browser_auth'")
+            raise ValueError(
+                "Incompatible options 'unauthenticated_push' and 'require_browser_auth'"
+            )
     if htdigest_file and not (require_browser_auth or use_smarthttp):
-        raise ValueError("'htdigest_file' set without 'use_smarthttp' or 'require_browser_auth'")
+        raise ValueError(
+            "'htdigest_file' set without 'use_smarthttp' or 'require_browser_auth'"
+        )
+    if not isinstance(repo_paths, dict):
+        # If repos is given as a flat list, put all repos under the "no namespace" namespace
+        repo_paths = {None: repo_paths}
 
     app = Klaus(
         repo_paths,
@@ -140,7 +172,10 @@ def make_app(repo_paths, site_name, use_smarthttp=False, htdigest_file=None,
     if use_smarthttp:
         # `path -> Repo` mapping for Dulwich's web support
         dulwich_backend = dulwich.server.DictBackend(
-            {'/'+name: repo for name, repo in app.valid_repos.items()}
+            {
+                "/" + namespaced_name: repo
+                for namespaced_name, repo in app.valid_repos.items()
+            }
         )
         # Dulwich takes care of all Git related requests/URLs
         # and passes through everything else to klaus
@@ -164,7 +199,9 @@ def make_app(repo_paths, site_name, use_smarthttp=False, htdigest_file=None,
         # Git will never call /<repo-name>/git-receive-pack if authentication
         # failed for /info/refs, but since it's used to upload stuff to the server
         # we must secure it anyway for security reasons.
-        PATTERN = r'^/[^/]+/(info/refs\?service=git-receive-pack|git-receive-pack)$'
+        PATTERN = (
+            r"^/(~[^/]+/)?[^/]+/(info/refs\?service=git-receive-pack|git-receive-pack)$"
+        )
         if unauthenticated_push:
             # DANGER ZONE: Don't require authentication for push'ing
             app.wsgi_app = dulwich_wrapped_app
@@ -193,8 +230,7 @@ def make_app(repo_paths, site_name, use_smarthttp=False, htdigest_file=None,
 
     if require_browser_auth:
         app.wsgi_app = httpauth.DigestFileHttpAuthMiddleware(
-            htdigest_file,
-            wsgi_app=app.wsgi_app
+            htdigest_file, wsgi_app=app.wsgi_app
         )
 
     return app
